import typing

from typing_extensions import Protocol, TypeAlias

_Address: TypeAlias = typing.Tuple[typing.Any, ...]


class SocketInterface(Protocol):

    def send(self, data: bytes) -> int: ...

    def sendall(self, data: bytes) -> None: ...

    def recv(self, bufsize: int) -> bytes: ...

    def bind(self, address: typing.Union[_Address, bytes]) -> None: ...

    def fileno(self) -> int: ...

    def settimeout(self, value: typing.Union[float, None]) -> None: ...

    def close(self) -> None: ...



import typing
from functools import wraps

from aioarp import get_ip, get_mac


def mock_get_ip(ip: str) -> typing.Callable[[], str]:
    @wraps(get_ip)
    def inner():
        return ip

    return inner


def mock_get_mac(mac: str) -> typing.Callable[[str], str]:
    @wraps(get_mac)
    def inner(interface: str) -> str:
        return mac

    return inner



from aioarp import ArpPacket, EthPacket, HardwareType, Opcode, Protocol


class TestEth:

    def test_eth_parse(self):
        eth_header = (
            b"\x11\x11\x11\x11\x11\x11\x11"
            b"\x11\x11\x11\x11\x11\x08\x06"
        )

        eth_packet = EthPacket.parse(frame=eth_header)
        assert eth_packet.sender_mac == '11:11:11:11:11:11'
        assert eth_packet.target_mac == '11:11:11:11:11:11'
        assert eth_packet.proto == Protocol.arp

    def test_build(self):
        assert EthPacket(
            sender_mac='11:11:11:11:11:11',
            target_mac='11:11:11:11:11:11',
            proto=Protocol.ip
        ).build_frame() == (
                   b"\x11\x11\x11\x11\x11\x11\x11"
                   b"\x11\x11\x11\x11\x11\x08\x00"
               )


class TestArp:

    def test_parse(self):
        arp_header = (
            b"\x00\x01\x08\x00\x06\x04"
            b"\x00\x01\x11\x11\x11\x11"
            b"\x11\x11\x7f\x00\x00\x01"
            b"\x11\x11\x11\x11\x11\x11"
            b"\x7f\x00\x00\x01"
        )
        arp_packet = ArpPacket.parse(frame=arp_header)
        assert arp_packet.hardware_type == HardwareType.ethernet
        assert arp_packet.protocol_type == Protocol.ip
        assert arp_packet.hardware_length == 6
        assert arp_packet.protocol_length == 4
        assert arp_packet.opcode == Opcode.request
        assert arp_packet.sender_mac == '11:11:11:11:11:11'
        assert arp_packet.sender_ip == '127.0.0.1'
        assert arp_packet.target_mac == '11:11:11:11:11:11'
        assert arp_packet.target_ip == '127.0.0.1'

    def test_build(self):
        assert ArpPacket(
            hardware_type=HardwareType.ethernet,
            protocol_type=Protocol.ip,
            sender_mac='11:11:11:11:11:11',
            sender_ip='127.0.0.1',
            target_mac='11:11:11:11:11:11',
            target_ip='127.0.0.1',
        ).build_frame() == (
            b"\x00\x01\x08\x00\x06\x04"
            b"\x00\x01\x11\x11\x11\x11"
            b"\x11\x11\x7f\x00\x00\x01"
            b"\x11\x11\x11\x11\x11\x11"
            b"\x7f\x00\x00\x01"
        )




import pytest

import aioarp
from aioarp._mock import mock_get_ip, mock_get_mac


@pytest.fixture(scope='function', autouse=True)
def mock_ip_and_mac(monkeypatch):
    monkeypatch.setattr(aioarp._client, 'get_ip', mock_get_ip('127.0.0.1'))
    monkeypatch.setattr(aioarp._client, 'get_mac', mock_get_mac('11:11:11:11:11:11'))
    yield


import re

SUBS = [
    ('async def async_send_arp', 'def sync_send_arp'),
    ('async def', 'def'),
    ('from .backends._async import AsyncStream', 'from .backends._sync import Stream'),
    ('AsyncStream', 'Stream'),
    ('await ', ''),
    ('async_send_arp', 'sync_send_arp'),
    ('aioarp.arequest', 'aioarp.request')
]

COMPILED_SUBS = [
    (re.compile(r'(^|\b)' + regex + r'($|\b)'), replaced)
    for regex, replaced in SUBS
]


def unasync_line(line):
    for regex, replaced in COMPILED_SUBS:
        line = re.sub(regex, replaced, line)
    return line


def unasync_file(async_path, sync_path):
    with (open(async_path, "r") as in_file, \
          open(sync_path, "w", newline="") as sync_file):
            for line in in_file.readlines():
                line = unasync_line(line)
                sync_file.write(line)

unasync_file("aioarp/_async.py", "aioarp/_sync.py")
unasync_file("tests/test_async.py", "tests/test_sync.py")



import tempfile
import typing

from typing_extensions import TypeAlias

_Address: TypeAlias = typing.Tuple[typing.Any, ...]


class MockSocket:

    def __init__(self, to_receive: bytes, delay: typing.Optional[float] = None):
        self._file = tempfile.TemporaryFile('wb+')
        self._to_receive = to_receive

    def send(self, data: bytes) -> int:
        self._file.write(data)
        return len(data)

    def sendall(self, data: bytes) -> None:
        self._file.write(data)

    def recv(self, max_bytes: typing.Optional[int] = None) -> bytes:
        return self._to_receive

    def bind(self, address: typing.Union[_Address, bytes]) -> None: ...
    def settimeout(self, value: typing.Union[float, None]) -> None: ...

    def fileno(self) -> int:
        return self._file.fileno()

    def close(self) -> None:
        self._file.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



from aioarp import enforce_ip, enforce_mac, parse_ip, parse_mac


def test_enforce_mac():
    encoded = enforce_mac('00:00:00:00:00:00')
    assert False not in [byte == 0 for byte in encoded]


def test_enforce_ip():
    encoded = enforce_ip('127.0.0.1')
    assert encoded == bytes([127, 0, 0, 1])


def test_parse_mac():
    mac = parse_mac(bytes([0, 0, 0, 0, 0, 0]))
    assert mac == '00:00:00:00:00:00'


def test_parse_ip():
    parsed = parse_ip(bytes([127, 0, 0, 1]))
    assert parsed == '127.0.0.1'



import select
import socket
import typing

import anyio

from aioarp import _exceptions as exc

from ._base import SocketInterface

__all__ = (
    'AsyncStream',
)

class AsyncStream:

    def __init__(self,
                 interface: str,
                 sock: typing.Optional[SocketInterface] = None
                 ):
        self.sock: SocketInterface
        if not sock:  # pragma: no cover
            self.sock = socket.socket(socket.PF_PACKET, socket.SOCK_RAW, socket.ntohs(0x0003))
        else:
            self.sock = sock
        self.sock.bind((interface, 0))

    async def receive_frame(self, timeout: float) -> bytes:
        frame: typing.Optional[bytes]
        with anyio.move_on_after(timeout):
            while True:
                r, _, _ = select.select([self.sock], [], [], 0.00001)
                if r:
                    frame = self.sock.recv(1123123)
                    break
                await anyio.sleep(0)  # pragma: no cover
        if not frame:   # pragma: no cover
            raise exc.ReadTimeoutError()
        return frame

    async def write_frame(self, frame: bytes, timeout: float) -> None:

        with anyio.move_on_after(timeout):
            while frame:
                _, w, _ = select.select([], [self.sock], [], 0.00001)

                nsent = self.sock.send(frame)
                frame = frame[nsent:]
                await anyio.sleep(0)
        if frame:  # pragma: no cover
            raise exc.WriteTimeoutError()

    def close(self) -> None:
        self.sock.close()

    def __enter__(self) -> "AsyncStream":
        return self

    def __exit__(self, exc_type: typing.Any, exc_val: typing.Any, exc_tb: typing.Any) -> None:
        self.close()



import time
import typing

from aioarp import _exceptions as exc
from aioarp._arp import ARP_HEADER_SIZE, ETHERNET_HEADER_SIZE, ArpPacket, EthPacket, Protocol
from aioarp._utils import is_valid_ipv4
from aioarp.defaults import DEFAULT_READ_TIMEOUT, DEFAULT_REPLY_MISSING_TIME, DEFAULT_WRITE_TIMEOUT

from ._backends._async import AsyncStream

__all__ = (
    'async_send_arp',
)


async def receive_arp(sock: AsyncStream, timeout: float) -> ArpPacket:
    start_time = time.time()
    while True:

        # Check if timeout was expired
        if time.time() - start_time > timeout:
            raise exc.NotFoundError()

        # Try to read frame
        try:
            frame = await sock.receive_frame(timeout=DEFAULT_READ_TIMEOUT)
        except Exception as e:  # pragma: no cover
            raise exc.NotFoundError() from e

        # Extract the ethernet header
        eth_header = frame[:ETHERNET_HEADER_SIZE]

        try:
            eth_packet = EthPacket.parse(eth_header)
            if eth_packet.proto != Protocol.arp:
                continue

            arp_response = ArpPacket.parse(
                frame[ETHERNET_HEADER_SIZE: ETHERNET_HEADER_SIZE + ARP_HEADER_SIZE])
            if is_valid_ipv4(arp_response.sender_ip):
                return arp_response
        except BaseException:  # pragma: no cover
            # TODO: catch concrete errors
            ...


async def async_send_arp(arp_packet: ArpPacket,
                         stream: AsyncStream,
                         timeout: typing.Optional[float] = None) -> ArpPacket:
    ethernet_packet = EthPacket(
        target_mac=arp_packet.target_mac,
        sender_mac=arp_packet.sender_mac,
        proto=Protocol.arp
    )

    try:
        frame_to_send = ethernet_packet.build_frame() + arp_packet.build_frame()
        await stream.write_frame(frame_to_send, timeout=DEFAULT_WRITE_TIMEOUT)
    except exc.WriteTimeoutError as e:  # pragma: no cover
        raise exc.NotFoundError from e

    return await receive_arp(stream, timeout or DEFAULT_REPLY_MISSING_TIME)



import socket
import typing

import aioarp
from aioarp._arp import ArpPacket, HardwareType, Protocol
from aioarp._async import async_send_arp
from aioarp._backends._async import AsyncStream
from aioarp._backends._base import SocketInterface
from aioarp._backends._sync import Stream
from aioarp._sync import sync_send_arp
from aioarp._utils import get_ip, get_mac, is_valid_ipv4

__all__ = (
    'build_arp_packet',
    'request',
    'arequest'
)


def build_arp_packet(
        interface: str,
        target_ip: str
) -> ArpPacket:
    if not is_valid_ipv4(target_ip):
        raise aioarp.InvalidIpError("Invalid IPv4 Address was received")

    hardware_type = HardwareType.ethernet
    protocol_type = Protocol.ip

    # TODO: catch interface not found error
    sender_mac = get_mac(interface)
    sender_ip = get_ip()
    target_mac = 'ff:ff:ff:ff:ff:ff'

    request_packet = ArpPacket(
        hardware_type,
        protocol_type,
        sender_mac,
        sender_ip,
        target_mac,
        target_ip,
    )
    return request_packet


def request(
        interface: str,
        target_ip: str,
        timeout: typing.Optional[float] = None,
        sock: typing.Optional[SocketInterface] = None
) -> ArpPacket:
    if not sock:  # pragma: no cover
        sock = socket.socket(socket.PF_PACKET, socket.SOCK_RAW, socket.ntohs(0x0003))
    with Stream(interface=interface,
                    sock=sock) as stream:
        request_packet = build_arp_packet(interface, target_ip)
        arp_response = sync_send_arp(request_packet, stream, timeout)
        return arp_response


async def arequest(
        interface: str,
        target_ip: str,
        timeout: typing.Optional[float] = None,
        sock: typing.Optional[SocketInterface] = None
) -> ArpPacket:
    if not sock:  # pragma: no cover
        sock = socket.socket(socket.PF_PACKET, socket.SOCK_RAW, socket.ntohs(0x0003))
    with AsyncStream(interface=interface,
                    sock=sock) as stream:
        request_packet = build_arp_packet(interface, target_ip)
        arp_response = await async_send_arp(request_packet, stream, timeout)
        return arp_response



import struct
import typing
from enum import Enum

from aioarp._utils import enforce_ip, enforce_mac, parse_ip, parse_mac

ETHERNET_HEADER_SIZE = 14
ARP_HEADER_SIZE = 28

__all__ = (
    'ETHERNET_HEADER_SIZE',
    'ARP_HEADER_SIZE',
    'HardwareType',
    'Protocol',
    'Opcode',
    'EthPacket',
    'ArpPacket',
    'size_of'
)

class HardwareType(Enum):
    ethernet = 1


class Protocol(Enum):
    ip = 0x0800
    arp = 0x0806


class Opcode(Enum):
    request = 1
    response = 2


class EthPacket:
    format = "!6s6sH"

    def __init__(self,
                 target_mac: str,
                 sender_mac: str,
                 proto: Protocol):
        self.target_mac = target_mac
        self.sender_mac = sender_mac
        self.proto = proto

    @classmethod
    def parse(cls, frame: bytes) -> "EthPacket":
        target_mac, sender_mac, proto = struct.unpack(
            '!6s6sH',
            frame
        )
        return cls(
            target_mac=parse_mac(target_mac),
            sender_mac=parse_mac(sender_mac),
            proto=Protocol(proto)
        )

    def build_frame(self) -> bytes:
        return struct.pack(
            self.format,
            enforce_mac(self.target_mac),
            enforce_mac(self.sender_mac),
            self.proto.value
        )

    def __repr__(self) -> str:
        return f"<EthPacket target={self.target_mac} source={self.sender_mac} proto={self.proto.name}>"


class ArpPacket:

    def __init__(self,
                 hardware_type: HardwareType,
                 protocol_type: Protocol,
                 sender_mac: str,
                 sender_ip: str,
                 target_mac: str,
                 target_ip: str,
                 opcode: Opcode = Opcode.request):
        self.hardware_type = hardware_type
        self.protocol_type = protocol_type
        self.opcode = opcode
        self.sender_mac = sender_mac
        self.sender_ip = sender_ip
        self.target_mac = target_mac
        self.target_ip = target_ip

    @property
    def hardware_length(self) -> int:
        return size_of(self.hardware_type)

    @property
    def protocol_length(self) -> int:
        return size_of(self.protocol_type)

    @classmethod
    def parse(cls, frame: bytes) -> "ArpPacket":
        hardware_format = '6s'
        protocol_format = '4s'
        packing_format = ''.join([
            "!"
            "H",  # hardware type
            "H",  # protocol type
            'B',  # hardware_length
            'B',  # protocol_length
            'H',  # opcode: Opcode = Opcode.request
            hardware_format,  # sender_mac
            protocol_format,  # sender_ip: int
            hardware_format,  # target_map: int
            protocol_format,  # target_ip: int
        ])

        (
            hardware_type,
            protocol_type,
            hardware_length,
            protocol_length,
            opcode,
            sender_mac,
            sender_ip,
            target_mac,
            target_ip
        ) = struct.unpack(
            packing_format,
            frame
        )
        return cls(
            hardware_type=HardwareType(hardware_type),
            protocol_type=Protocol(protocol_type),
            opcode=Opcode(opcode),
            sender_mac=parse_mac(sender_mac),
            sender_ip=parse_ip(sender_ip),
            target_mac=parse_mac(target_mac),
            target_ip=parse_ip(target_ip)
        )

    def build_frame(self) -> bytes:
        hardware_format = str(self.hardware_length) + 's'
        protocol_format = str(self.protocol_length) + 's'
        packing_format = ''.join([
            "!"
            "H",  # hardware type
            "H",  # protocol type
            'B',  # hardware_length
            'B',  # protocol_length
            'H',  # opcode: Opcode = Opcode.request
            hardware_format,  # sender_mac
            protocol_format,  # sender_ip: int
            hardware_format,  # target_map: int
            protocol_format,  # target_ip: int
        ])

        return struct.pack(
            packing_format,
            self.hardware_type.value,
            self.protocol_type.value,
            self.hardware_length,
            self.protocol_length,
            self.opcode.value,
            enforce_mac(self.sender_mac),
            enforce_ip(self.sender_ip),
            enforce_mac(self.target_mac),
            enforce_ip(self.target_ip),
        )

    def __repr__(self) -> str:
        return f"<ArpPacket target_ip={self.target_ip}>"


def size_of(type: typing.Union[HardwareType, Protocol]) -> int:
    if type is Protocol.ip:
        return 4
    elif type is HardwareType.ethernet:
        return 6
    else:
        raise NotImplementedError()




import pytest

import aioarp
from aioarp import HardwareType, Protocol, build_arp_packet


def test_invalid_ip():
    with pytest.raises(aioarp.InvalidIpError):
        aioarp.build_arp_packet('test', '10.10.10.10.10')


def test_build_packet(monkeypatch):
    packet = build_arp_packet('test', '192.168.0.85')

    assert packet.sender_mac == '11:11:11:11:11:11'
    assert packet.sender_ip == '127.0.0.1'
    assert packet.hardware_type == HardwareType.ethernet
    assert packet.protocol_type == Protocol.ip
    assert packet.target_mac == 'ff:ff:ff:ff:ff:ff'
    assert packet.target_ip == '192.168.0.85'




__all__ = (
    'AioArpError',
    'InvalidIpError',
    'NotFoundError',
    'TimeoutError',
    'ReadTimeoutError'
)

class AioArpError(Exception):
    ...

class InvalidIpError(AioArpError):
    ...

class NotFoundError(AioArpError):
    ...

class TimeoutError(AioArpError):
    ...

class ReadTimeoutError(TimeoutError):
    ...

class WriteTimeoutError(TimeoutError):
    ...



import pytest

import aioarp
from aioarp._backends._mock import MockSocket

eth_header_arp = (
    # Eth Header
    b"\x11\x11\x11\x11\x11\x11\x11"
    b"\x11\x11\x11\x11\x11\x08\x06"
)

eth_header_no_arp = (
    # Eth Header
    b"\x11\x11\x11\x11\x11\x11\x11"
    b"\x11\x11\x11\x11\x11\x08\x00"
)

arp_header = (
    # Arp Header
    b"\x00\x01\x08\x00\x06\x04"
    b"\x00\x01\x11\x11\x11\x11"
    b"\x11\x11\x7f\x00\x00\x01"
    b"\x11\x11\x11\x11\x11\x11"
    b"\x7f\x00\x00\x01"
)


@pytest.mark.anyio
async def test_async_send_arp():
    msocket = MockSocket(eth_header_arp + arp_header)
    response = await aioarp.arequest('null', '100.100.100.100', sock=msocket)
    assert response.sender_mac == '11:11:11:11:11:11'


@pytest.mark.anyio
async def test_async_send_arp_timeout():
    with MockSocket(eth_header_no_arp + arp_header) as msocket:
        with pytest.raises(aioarp.NotFoundError):
            await aioarp.arequest('null', '100.100.100.100', sock=msocket, timeout=0.5)



import fcntl
import ipaddress
import socket
import struct
import typing

__all__ = (
    'is_valid_ipv4',
    'get_mac',
    'get_ip',
    'enforce_mac',
    'enforce_ip',
    'parse_mac',
    'parse_ip'
)

OUR_MAC = None
OUR_IP = None


def is_valid_ipv4(ip: str) -> bool:
    try:
        ipaddress.IPv4Address(ip)
        return True
    except ipaddress.AddressValueError:
        return False


def get_mac(interface: str) -> str:  # pragma: no cover
    global OUR_MAC
    if OUR_MAC:
        return OUR_MAC
    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
        info = fcntl.ioctl(s.fileno(), 0x8927, struct.pack('256s', bytes(interface, 'utf-8')[:15]))
        mac = ':'.join('%02x' % b for b in info[18:24])
        OUR_MAC = mac
        return OUR_MAC


def get_ip() -> str:  # pragma: no cover
    global OUR_IP
    if OUR_IP:
        return OUR_IP
    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
        s.connect(("1.1.1.1", 80))
        ip = typing.cast(str, s.getsockname()[0])
        OUR_IP = ip
        return OUR_IP


def enforce_mac(mac: str) -> bytes:
    mac_bytes = []
    for b in mac.split(':'):
        mac_bytes.append(int(b, 16))
    return bytes(mac_bytes)


def enforce_ip(ip: str) -> bytes:
    ip_bytes = []
    for b in ip.split('.'):
        ip_bytes.append(int(b))
    return bytes(ip_bytes)


def parse_mac(mac: bytes) -> str:
    mac_parts = []
    for b in mac:
        hex = f'{b:x}'
        if len(hex) == 1:
            hex = '0' + hex
        mac_parts.append(hex)
    return ':'.join(mac_parts)


def parse_ip(ip: bytes) -> str:
    ip_parts = []
    for b in ip:
        ip_parts.append(str(b))
    return '.'.join(ip_parts)



import socket
import typing

from aioarp import _exceptions as exc
from aioarp._backends._base import SocketInterface

__all__ = (
    'Stream',
)
class Stream:

    def __init__(self,
                 interface: str,
                 sock: typing.Optional[SocketInterface] = None
                 ):
        self.sock: SocketInterface
        if not sock:  # pragma: no cover
            self.sock = socket.socket(socket.PF_PACKET, socket.SOCK_RAW, socket.ntohs(0x0003))
        else:
            self.sock = sock
        self.sock.bind((interface, 0))

    def receive_frame(self, timeout: float) -> bytes:
        self.sock.settimeout(timeout)
        try:
            frame = self.sock.recv(1123123)
        except socket.timeout:  # pragma: no cover
            raise exc.ReadTimeoutError()
        return frame

    def write_frame(self, frame: bytes, timeout: float) -> None:
        self.sock.settimeout(timeout)
        try:
            self.sock.sendall(frame)
        except socket.timeout:  # pragma: no cover
            raise exc.WriteTimeoutError()

    def __enter__(self) -> 'Stream':
        return self

    def __exit__(self, exc_type: typing.Any, exc_val: typing.Any, exc_tb: typing.Any) -> None:
        self.sock.close()
